@@ -26,6 +26,7 @@
 #include <glib.h>
 
 #include <telepathy-glib/debug.h>
+#include <telepathy-glib/debug-sender.h>
 
 #include "debug.h"
 
@@ -36,6 +37,7 @@ static const GDebugKey tpsip_debug_keys[] = {
   { "connection",    TPSIP_DEBUG_CONNECTION },
   { "im",            TPSIP_DEBUG_IM },
   { "events",        TPSIP_DEBUG_EVENTS },
+  { "sofia",         TPSIP_DEBUG_SOFIA },
 };
 
 void tpsip_debug_set_flags_from_env ()
@@ -66,17 +68,78 @@ gboolean tpsip_debug_flag_is_set (TpsipDebugFlags flag)
   return (flag & tpsip_debug_flags) ? TRUE : FALSE;
 }
 
-void tpsip_debug (TpsipDebugFlags flag,
-                  const gchar *format,
-                  ...)
+static GHashTable *flag_to_domains = NULL;
+
+static const gchar *
+debug_flag_to_domain (TpsipDebugFlags flag)
 {
-  if (flag & tpsip_debug_flags)
+  if (G_UNLIKELY (flag_to_domains == NULL))
     {
-      va_list args;
-      va_start (args, format);
-      g_logv (G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, format, args);
-      va_end (args);
+      guint i;
+
+      flag_to_domains = g_hash_table_new_full (g_direct_hash, g_direct_equal,
+          NULL, g_free);
+
+      for (i = 0; tpsip_debug_keys[i].value; i++)
+        {
+          GDebugKey key = (GDebugKey) tpsip_debug_keys[i];
+          gchar *val;
+
+          val = g_strdup_printf ("%s/%s", G_LOG_DOMAIN, key.key);
+          g_hash_table_insert (flag_to_domains,
+              GUINT_TO_POINTER (key.value), val);
+        }
     }
+
+  return g_hash_table_lookup (flag_to_domains, GUINT_TO_POINTER (flag));
+}
+
+void
+tpsip_debug_free (void)
+{
+  if (flag_to_domains == NULL)
+    return;
+
+  g_hash_table_destroy (flag_to_domains);
+  flag_to_domains = NULL;
+}
+
+static void
+log_to_debug_sender (TpsipDebugFlags flag,
+                     GLogLevelFlags level,
+                     const gchar *message)
+{
+  TpDebugSender *dbg;
+  GTimeVal now;
+
+  dbg = tp_debug_sender_dup ();
+
+  g_get_current_time (&now);
+
+  tp_debug_sender_add_message (dbg, &now, debug_flag_to_domain (flag),
+      level, message);
+
+  g_object_unref (dbg);
+}
+
+void tpsip_log (TpsipDebugFlags flag,
+                GLogLevelFlags level,
+                const gchar *format,
+                ...)
+{
+  gchar *message;
+  va_list args;
+
+  va_start (args, format);
+  message = g_strdup_vprintf (format, args);
+  va_end (args);
+
+  log_to_debug_sender (flag, level, message);
+
+  if (flag & tpsip_debug_flags)
+    g_log (G_LOG_DOMAIN, level, "%s", message);
+
+  g_free (message);
 }
 
 #endif /* ENABLE_DEBUG */
