@@ -31,26 +31,19 @@
 
 #include <dbus/dbus-protocol.h>
 
+#include <telepathy-glib/debug-sender.h>
 #include <telepathy-glib/errors.h>
 #include <telepathy-glib/svc-connection-manager.h>
 
 #include <tpsip/sofia-decls.h>
 #include <sofia-sip/su_glib.h>
 
-#ifdef HAVE_LIBIPHB
-#include <libiphb.h>
-#endif
-
 #include "sip-connection-manager.h"
 #include "sip-connection.h"
 
 #define DEBUG_FLAG TPSIP_DEBUG_CONNECTION
 #include "debug.h"
 
-/* Time bounds for heartbeat, in seconds */
-#define TPSIP_HEARTBEAT_MIN 0
-#define TPSIP_HEARTBEAT_MAX 30
-
 
 G_DEFINE_TYPE(TpsipConnectionManager, tpsip_connection_manager,
     TP_TYPE_BASE_CONNECTION_MANAGER)
@@ -73,6 +66,7 @@ typedef struct {
     gboolean discover_stun;
     gchar *stun_server;
     guint stun_port;
+    gboolean immutable_streams;
     gchar *local_ip_address;
     guint local_port;
     gchar *extra_auth_user;
@@ -122,6 +116,7 @@ enum {
     TPSIP_CONN_PARAM_DISCOVER_STUN,
     TPSIP_CONN_PARAM_STUN_SERVER,
     TPSIP_CONN_PARAM_STUN_PORT,
+    TPSIP_CONN_PARAM_IMMUTABLE_STREAMS,
     TPSIP_CONN_PARAM_LOCAL_IP_ADDRESS,
     TPSIP_CONN_PARAM_LOCAL_PORT,
     TPSIP_CONN_PARAM_EXTRA_AUTH_USER,
@@ -139,9 +134,7 @@ static const TpCMParamSpec tpsip_params[] = {
       0, NULL, G_STRUCT_OFFSET (TpsipConnParams, auth_user) },
     /* Password */
     { "password", DBUS_TYPE_STRING_AS_STRING, G_TYPE_STRING,
-      0, /* according to the .manager file this is 
-      TP_CONN_MGR_PARAM_FLAG_REQUIRED | TP_CONN_MGR_PARAM_FLAG_REGISTER,
-      but in the code this is not the case */
+      TP_CONN_MGR_PARAM_FLAG_SECRET,
       NULL, G_STRUCT_OFFSET (TpsipConnParams, password) },
     /* Display name for self */
     { "alias", DBUS_TYPE_STRING_AS_STRING, G_TYPE_STRING, 0, NULL,
@@ -172,10 +165,12 @@ static const TpCMParamSpec tpsip_params[] = {
       G_STRUCT_OFFSET (TpsipConnParams, discover_binding) },
     /* Mechanism used for connection keepalive maintenance */
     { "keepalive-mechanism", DBUS_TYPE_STRING_AS_STRING, G_TYPE_STRING,
-      0, NULL, G_STRUCT_OFFSET (TpsipConnParams, keepalive_mechanism) },
+      TP_CONN_MGR_PARAM_FLAG_HAS_DEFAULT, "auto",
+      G_STRUCT_OFFSET (TpsipConnParams, keepalive_mechanism) },
     /* Keep-alive interval */
     { "keepalive-interval", DBUS_TYPE_UINT32_AS_STRING, G_TYPE_UINT,
-      0, NULL, G_STRUCT_OFFSET (TpsipConnParams, keepalive_interval) },
+      TP_CONN_MGR_PARAM_FLAG_HAS_DEFAULT, GUINT_TO_POINTER(0),
+      G_STRUCT_OFFSET (TpsipConnParams, keepalive_interval) },
     /* Use SRV DNS lookup to discover STUN server */
     { "discover-stun", DBUS_TYPE_BOOLEAN_AS_STRING, G_TYPE_BOOLEAN,
       TP_CONN_MGR_PARAM_FLAG_HAS_DEFAULT, GUINT_TO_POINTER(TRUE),
@@ -188,6 +183,10 @@ static const TpCMParamSpec tpsip_params[] = {
       TP_CONN_MGR_PARAM_FLAG_HAS_DEFAULT,
       GUINT_TO_POINTER(TPSIP_DEFAULT_STUN_PORT),
       G_STRUCT_OFFSET (TpsipConnParams, stun_port) },
+    /* If the session content cannot be modified once initially set up */
+    { "immutable-streams", DBUS_TYPE_BOOLEAN_AS_STRING, G_TYPE_BOOLEAN,
+      TP_CONN_MGR_PARAM_FLAG_HAS_DEFAULT, GUINT_TO_POINTER(FALSE),
+      G_STRUCT_OFFSET (TpsipConnParams, immutable_streams) },
     /* Local IP address to use, workaround purposes only */
     { "local-ip-address", DBUS_TYPE_STRING_AS_STRING, G_TYPE_STRING,
       0, NULL, G_STRUCT_OFFSET (TpsipConnParams, local_ip_address) },
@@ -198,7 +197,8 @@ static const TpCMParamSpec tpsip_params[] = {
     { "extra-auth-user", DBUS_TYPE_STRING_AS_STRING, G_TYPE_STRING,
       0, NULL, G_STRUCT_OFFSET (TpsipConnParams, extra_auth_user) },
     { "extra-auth-password", DBUS_TYPE_STRING_AS_STRING, G_TYPE_STRING,
-      0, NULL, G_STRUCT_OFFSET (TpsipConnParams, extra_auth_password) },
+      TP_CONN_MGR_PARAM_FLAG_SECRET,
+      NULL, G_STRUCT_OFFSET (TpsipConnParams, extra_auth_password) },
     { NULL, NULL, 0, 0, NULL, 0 }
 };
 
@@ -210,103 +210,11 @@ const TpCMProtocolSpec tpsip_protocols[] = {
 struct _TpsipConnectionManagerPrivate
 {
   su_root_t *sofia_root;
-
-#ifdef HAVE_LIBIPHB
-  iphb_t    heartbeat;
-  su_wait_t heartbeat_wait[1];
-  int       heartbeat_wait_id;
-#endif
+  TpDebugSender *debug_sender;
 };
 
 #define TPSIP_CONNECTION_MANAGER_GET_PRIVATE(obj) ((obj)->priv)
 
-#ifdef HAVE_LIBIPHB
-
-static void heartbeat_shutdown (TpsipConnectionManager *self);
-
-static int
-heartbeat_wakeup (TpsipConnectionManager *self,
-                  su_wait_t *wait,
-                  void *user_data)
-{
-  TpsipConnectionManagerPrivate *priv = TPSIP_CONNECTION_MANAGER_GET_PRIVATE (self);
-
-  DEBUG("tick");
-
-  g_assert (priv->heartbeat != NULL);
-
-  if ((wait->revents & (SU_WAIT_HUP | SU_WAIT_ERR)) != 0)
-    {
-      g_warning ("heartbeat descriptor invalidated prematurely with event mask %hd", wait->revents);
-      heartbeat_shutdown (self);
-      return 0;
-    }
-
-  iphb_wait (priv->heartbeat, TPSIP_HEARTBEAT_MIN, TPSIP_HEARTBEAT_MAX, 0);
-
-  return 0;
-}
-
-#endif /* HAVE_LIBIPHB */
-
-static void
-heartbeat_init (TpsipConnectionManager *self)
-{
-#ifdef HAVE_LIBIPHB
-  TpsipConnectionManagerPrivate *priv = TPSIP_CONNECTION_MANAGER_GET_PRIVATE (self);
-  int wait_id;
-  int reference_interval = 0;
-
-  su_root_set_max_defer (priv->sofia_root, TPSIP_HEARTBEAT_MAX * 1000L + 50L);
-
-  priv->heartbeat = iphb_open (&reference_interval);
-
-  if (priv->heartbeat == NULL)
-    {
-      g_warning ("opening IP heartbeat failed: %s", strerror (errno));
-      return;
-    }
-
-  DEBUG("heartbeat opened with reference interval %d", reference_interval);
-
-  su_wait_init (priv->heartbeat_wait);
-  if (su_wait_create (priv->heartbeat_wait,
-                      iphb_get_fd (priv->heartbeat),
-                      SU_WAIT_IN) != 0)
-    g_critical ("could not create a wait object");
-
-  wait_id = su_root_register (priv->sofia_root,
-      priv->heartbeat_wait, heartbeat_wakeup, NULL, 0);
-
-  g_return_if_fail (wait_id > 0);
-  priv->heartbeat_wait_id = wait_id;
-
-  /* Prime the heartbeat for the first time.
-   * The correct sequence is iphb_wait() -> poll */
-  iphb_wait (priv->heartbeat, TPSIP_HEARTBEAT_MIN, TPSIP_HEARTBEAT_MAX, 0);
-
-#endif /* HAVE_LIBIPHB */
-}
-
-static void
-heartbeat_shutdown (TpsipConnectionManager *self)
-{
-#ifdef HAVE_LIBIPHB
-  TpsipConnectionManagerPrivate *priv = TPSIP_CONNECTION_MANAGER_GET_PRIVATE (self);
-
-  if (priv->heartbeat_wait_id == 0)
-    return;
-
-  su_root_deregister (priv->sofia_root, priv->heartbeat_wait_id);
-  priv->heartbeat_wait_id = 0;
-
-  su_wait_destroy (priv->heartbeat_wait);
-
-  iphb_close (priv->heartbeat);
-  priv->heartbeat = NULL;
-#endif /* HAVE_LIBIPHB */
-}
-
 static void
 tpsip_connection_manager_init (TpsipConnectionManager *obj)
 {
@@ -321,7 +229,12 @@ tpsip_connection_manager_init (TpsipConnectionManager *obj)
   source = su_root_gsource(priv->sofia_root);
   g_source_attach(source, NULL);
 
-  heartbeat_init (obj);
+  priv->debug_sender = tp_debug_sender_dup ();
+  g_log_set_default_handler (tp_debug_sender_log_handler, G_LOG_DOMAIN);
+
+#ifdef HAVE_LIBIPHB
+  su_root_set_max_defer (priv->sofia_root, TPSIP_DEFER_TIMEOUT * 1000L);
+#endif
 }
 
 static void tpsip_connection_manager_finalize (GObject *object);
@@ -352,12 +265,18 @@ tpsip_connection_manager_finalize (GObject *object)
   TpsipConnectionManagerPrivate *priv = TPSIP_CONNECTION_MANAGER_GET_PRIVATE (self);
   GSource *source;
 
-  heartbeat_shutdown (self);
-
   source = su_root_gsource(priv->sofia_root);
   g_source_destroy(source);
   su_root_destroy(priv->sofia_root);
 
+  if (priv->debug_sender != NULL)
+    {
+      g_object_unref (priv->debug_sender);
+      priv->debug_sender = NULL;
+    }
+
+  tpsip_debug_free ();
+
   G_OBJECT_CLASS (tpsip_connection_manager_parent_class)->finalize (object);
 }
 
@@ -457,7 +376,7 @@ priv_parse_keepalive (const gchar *str)
   if (strcmp (str, "off") == 0)
     return TPSIP_CONNECTION_KEEPALIVE_NONE;
 
-  g_warning ("unsupported keepalive-method value \"%s\", falling back to auto", str);
+  WARNING ("unsupported keepalive-method value \"%s\", falling back to auto", str);
   return TPSIP_CONNECTION_KEEPALIVE_AUTO;
 }
 
@@ -583,6 +502,9 @@ tpsip_connection_manager_new_connection (TpBaseConnectionManager *base,
   SET_PROPERTY_IF_PARAM_SET ("stun-port", TPSIP_CONN_PARAM_STUN_PORT,
       params->stun_port);
 
+  SET_PROPERTY_IF_PARAM_SET ("immutable-streams", TPSIP_CONN_PARAM_IMMUTABLE_STREAMS,
+      params->immutable_streams);
+
   SET_PROPERTY_IF_PARAM_SET ("keepalive-interval",
       TPSIP_CONN_PARAM_KEEPALIVE_INTERVAL, params->keepalive_interval);
 
