@@ -1,7 +1,7 @@
 /*
  * sip-media-channel.c - Source for TpsipMediaChannel
  * Copyright (C) 2005-2008 Collabora Ltd.
- * Copyright (C) 2005-2008 Nokia Corporation
+ * Copyright (C) 2005-2010 Nokia Corporation
  *   @author Kai Vehmanen <first.surname@nokia.com>
  *   @author Mikhail Zabaluev <mikhail.zabaluev@nokia.com>
  *
@@ -43,6 +43,11 @@
 #include "sip-connection-helpers.h"
 #include "sip-media-session.h"
 
+#define TPSIP_CHANNEL_CALL_STATE_PROCEEDING_MASK \
+    (TP_CHANNEL_CALL_STATE_RINGING | \
+     TP_CHANNEL_CALL_STATE_QUEUED | \
+     TP_CHANNEL_CALL_STATE_IN_PROGRESS)
+
 static void event_target_init (gpointer, gpointer);
 static void channel_iface_init (gpointer, gpointer);
 static void media_signalling_iface_init (gpointer, gpointer);
@@ -95,11 +100,13 @@ enum
   PROP_TARGET_ID,
   PROP_INITIATOR,
   PROP_INITIATOR_ID,
-  PROP_PEER,
   PROP_REQUESTED,
   PROP_INTERFACES,
   PROP_CHANNEL_DESTROYED,
   PROP_CHANNEL_PROPERTIES,
+  PROP_INITIAL_AUDIO,
+  PROP_INITIAL_VIDEO,
+  PROP_IMMUTABLE_STREAMS,
   /* Telepathy properties (see below too) */
   PROP_NAT_TRAVERSAL,
   PROP_STUN_SERVER,
@@ -123,6 +130,16 @@ static const TpPropertySignature media_channel_property_signatures[NUM_TP_PROPS]
     { "stun-port",              G_TYPE_UINT },
 };
 
+/* signals */
+enum
+{
+  SIG_INCOMING_CALL,
+  NUM_SIGNALS
+};
+
+static guint signals[NUM_SIGNALS] = { 0 };
+
+
 /* private structure */
 typedef struct _TpsipMediaChannelPrivate TpsipMediaChannelPrivate;
 
@@ -134,7 +151,12 @@ struct _TpsipMediaChannelPrivate
   TpHandle handle;
   TpHandle initiator;
   GHashTable *call_states;
+  gchar *stun_server;
+  guint stun_port;
 
+  gboolean initial_audio;
+  gboolean initial_video;
+  gboolean immutable_streams;
   gboolean closed;
   gboolean dispose_has_run;
 };
@@ -249,12 +271,24 @@ tpsip_media_channel_class_init (TpsipMediaChannelClass *klass)
       { "Requested", "requested", NULL },
       { NULL }
   };
+  static TpDBusPropertiesMixinPropImpl streamed_media_props[] = {
+      { "InitialAudio", "initial-audio", NULL },
+      { "InitialVideo", "initial-video", NULL },
+      { "ImmutableStreams", "immutable-streams", NULL },
+      { NULL }
+  };
+
   static TpDBusPropertiesMixinIfaceImpl prop_interfaces[] = {
       { TP_IFACE_CHANNEL,
         tp_dbus_properties_mixin_getter_gobject_properties,
         NULL,
         channel_props,
       },
+      { TP_IFACE_CHANNEL_TYPE_STREAMED_MEDIA,
+        tp_dbus_properties_mixin_getter_gobject_properties,
+        NULL,
+        streamed_media_props,
+      },
       { NULL }
   };
 
@@ -337,6 +371,36 @@ tpsip_media_channel_class_init (TpsipMediaChannelClass *klass)
       G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class, PROP_REQUESTED, param_spec);
 
+  param_spec = g_param_spec_boolean ("initial-audio", "InitialAudio",
+      "Whether the channel initially contained an audio stream",
+      FALSE,
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_INITIAL_AUDIO,
+      param_spec);
+
+  param_spec = g_param_spec_boolean ("initial-video", "InitialVideo",
+      "Whether the channel initially contained a video stream",
+      FALSE,
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_INITIAL_VIDEO,
+      param_spec);
+
+  param_spec = g_param_spec_boolean ("immutable-streams", "ImmutableStreams",
+      "Whether the set of streams on this channel are fixed once requested",
+      FALSE,
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_IMMUTABLE_STREAMS,
+      param_spec);
+
+  signals[SIG_INCOMING_CALL] =
+      g_signal_new ("incoming-call",
+                    G_OBJECT_CLASS_TYPE (klass),
+                    G_SIGNAL_RUN_LAST,
+                    0,
+                    NULL, NULL,
+                    g_cclosure_marshal_VOID__VOID,
+                    G_TYPE_NONE, 0);
+
   tp_properties_mixin_class_init (object_class,
       G_STRUCT_OFFSET (TpsipMediaChannelClass, properties_class),
       media_channel_property_signatures, NUM_TP_PROPS, NULL);
@@ -395,20 +459,6 @@ tpsip_media_channel_get_property (GObject    *object,
       else
         g_value_set_static_string (value, "");
       break;
-    case PROP_PEER:
-      {
-        TpHandle peer = 0;
-
-        if (priv->handle != 0)
-          peer = priv->handle;
-        else if (priv->session != NULL)
-          g_object_get (priv->session,
-              "peer", &peer,
-              NULL);
-
-        g_value_set_uint (value, peer);
-        break;
-      }
     case PROP_INITIATOR:
       g_value_set_uint (value, priv->initiator);
       break;
@@ -440,8 +490,26 @@ tpsip_media_channel_get_property (GObject    *object,
               TP_IFACE_CHANNEL, "InitiatorID",
               TP_IFACE_CHANNEL, "Requested",
               TP_IFACE_CHANNEL, "Interfaces",
+              TP_IFACE_CHANNEL_TYPE_STREAMED_MEDIA, "InitialAudio",
+              TP_IFACE_CHANNEL_TYPE_STREAMED_MEDIA, "InitialVideo",
+              TP_IFACE_CHANNEL_TYPE_STREAMED_MEDIA, "ImmutableStreams",
               NULL));
       break;
+    case PROP_INITIAL_AUDIO:
+      g_value_set_boolean (value, priv->initial_audio);
+      break;
+    case PROP_INITIAL_VIDEO:
+      g_value_set_boolean (value, priv->initial_video);
+      break;
+    case PROP_IMMUTABLE_STREAMS:
+      g_value_set_boolean (value, priv->immutable_streams);
+      break;
+    case PROP_STUN_SERVER:
+      g_value_set_string (value, priv->stun_server);
+      break;
+    case PROP_STUN_PORT:
+      g_value_set_uint (value, priv->stun_port);
+      break;
     default:
       /* Some properties live in the mixin */
       {
@@ -469,6 +537,32 @@ tpsip_media_channel_get_property (GObject    *object,
   }
 }
 
+static gboolean
+tpsip_media_channel_set_tp_property (TpsipMediaChannel *chan,
+                                     const GValue *value,
+                                     GParamSpec   *pspec)
+{
+  GObject *obj = (GObject *) chan;
+  const gchar *param_name = g_param_spec_get_name (pspec);
+  guint tp_property_id;
+
+  if (G_LIKELY (tp_properties_mixin_has_property (obj, param_name,
+                  &tp_property_id)))
+    {
+      tp_properties_mixin_change_value (obj, tp_property_id,
+          value, NULL);
+      tp_properties_mixin_change_flags (obj, tp_property_id,
+          TP_PROPERTY_FLAG_READ, 0, NULL);
+      return TRUE;
+    }
+  else
+    {
+      WARNING("Telepathy property '%s' is not defined for media channels",
+              param_name);
+      return FALSE;
+    }
+}
+
 static void
 tpsip_media_channel_set_property (GObject     *object,
 				guint        property_id,
@@ -502,22 +596,29 @@ tpsip_media_channel_set_property (GObject     *object,
       /* similarly we can't ref this yet */
       priv->initiator = g_value_get_uint (value);
       break;
+    case PROP_INITIAL_AUDIO:
+      priv->initial_audio = g_value_get_boolean (value);
+      break;
+    case PROP_INITIAL_VIDEO:
+      priv->initial_video = g_value_get_boolean (value);
+      break;
+    case PROP_IMMUTABLE_STREAMS:
+      priv->immutable_streams = g_value_get_boolean (value);
+      break;
+    case PROP_STUN_SERVER:
+      priv->stun_server = g_value_dup_string (value);
+      /* Also expose as a legacy Telepathy property */
+      tpsip_media_channel_set_tp_property (chan, value, pspec);
+      break;
+    case PROP_STUN_PORT:
+      priv->stun_port = g_value_get_uint (value);
+      /* Also expose as a legacy Telepathy property */
+      tpsip_media_channel_set_tp_property (chan, value, pspec);
+      break;
     default:
       /* some properties live in the mixin */
-      {
-        const gchar *param_name = g_param_spec_get_name (pspec);
-        guint tp_property_id;
-
-        if (G_LIKELY (tp_properties_mixin_has_property (object, param_name,
-                        &tp_property_id)))
-          {
-            tp_properties_mixin_change_value (object, tp_property_id,
-                value, NULL);
-            tp_properties_mixin_change_flags (object, tp_property_id,
-                TP_PROPERTY_FLAG_READ, 0, NULL);
-            return;
-          }
-      }
+      if (tpsip_media_channel_set_tp_property (chan, value, pspec))
+        return;
 
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
       break;
@@ -563,6 +664,8 @@ tpsip_media_channel_finalize (GObject *object)
 
   g_hash_table_destroy (priv->call_states);
 
+  g_free (priv->stun_server);
+
   g_free (priv->object_path);
 
   tp_group_mixin_finalize (object);
@@ -779,7 +882,12 @@ tpsip_media_channel_remove_streams (TpSvcChannelTypeStreamedMedia *iface,
 
   priv = TPSIP_MEDIA_CHANNEL_GET_PRIVATE (self);
 
-  if (priv->session != NULL)
+  if (priv->immutable_streams)
+    {
+      error = g_error_new (TP_ERRORS, TP_ERROR_NOT_IMPLEMENTED,
+          "Cannot remove streams from the existing channel");
+    }
+  else if (priv->session != NULL)
     {
        tpsip_media_session_remove_streams(priv->session,
                                         streams,
@@ -819,6 +927,14 @@ tpsip_media_channel_request_stream_direction (TpSvcChannelTypeStreamedMedia *ifa
 
   priv = TPSIP_MEDIA_CHANNEL_GET_PRIVATE (self);
 
+  if (priv->immutable_streams)
+    {
+      GError e = { TP_ERRORS, TP_ERROR_NOT_AVAILABLE,
+          "Cannot change directions on an immutable channel" };
+      dbus_g_method_return_error (context, &e);
+      return;
+    }
+
   if (priv->session != NULL)
     {
       tpsip_media_session_request_stream_direction (priv->session,
@@ -866,6 +982,14 @@ tpsip_media_channel_request_streams (TpSvcChannelTypeStreamedMedia *iface,
 
   priv = TPSIP_MEDIA_CHANNEL_GET_PRIVATE (self);
 
+  if (priv->immutable_streams)
+    {
+      GError e = { TP_ERRORS, TP_ERROR_NOT_AVAILABLE,
+          "Cannot add streams to the immutable channel" };
+      dbus_g_method_return_error (context, &e);
+      return;
+    }
+
   contact_repo = tp_base_connection_get_handles (
       (TpBaseConnection *)(priv->conn), TP_HANDLE_TYPE_CONTACT);
 
@@ -901,6 +1025,30 @@ tpsip_media_channel_request_streams (TpSvcChannelTypeStreamedMedia *iface,
  * Set: sip-media-channel API towards sip-connection
  ***********************************************************************/
 
+void
+tpsip_media_channel_create_initial_streams (TpsipMediaChannel *self)
+{
+  TpsipMediaChannelPrivate *priv = TPSIP_MEDIA_CHANNEL_GET_PRIVATE (self);
+
+  g_assert (priv->initiator != priv->handle);
+
+  priv_outbound_call (self, priv->handle);
+
+  g_assert (priv->session != NULL);
+
+  if (priv->initial_audio)
+    tpsip_media_session_add_stream (priv->session,
+        TP_MEDIA_STREAM_TYPE_AUDIO,
+        TP_MEDIA_STREAM_DIRECTION_BIDIRECTIONAL,
+        TRUE);
+
+  if (priv->initial_video)
+    tpsip_media_session_add_stream (priv->session,
+        TP_MEDIA_STREAM_TYPE_VIDEO,
+        TP_MEDIA_STREAM_DIRECTION_BIDIRECTIONAL,
+        TRUE);
+}
+
 /**
  * Handle an incoming INVITE, normally called just after the channel
  * has been created with initiator handle of the sender.
@@ -976,7 +1124,7 @@ tpsip_media_channel_peer_error (TpsipMediaChannel *self,
     case 404:
     case 480:
       reason = (tpsip_media_channel_get_call_state (self, peer)
-             & (TP_CHANNEL_CALL_STATE_RINGING | TP_CHANNEL_CALL_STATE_QUEUED))
+             & TPSIP_CHANNEL_CALL_STATE_PROCEEDING_MASK)
           ? TP_CHANNEL_GROUP_CHANGE_REASON_NO_ANSWER
           : TP_CHANNEL_GROUP_CHANGE_REASON_OFFLINE;
       break;
@@ -1137,6 +1285,34 @@ priv_nua_i_cancel_cb (TpsipMediaChannel *self,
   return TRUE;
 }
 
+static void
+priv_initial_media_properties_from_sdp (TpsipMediaChannel *self,
+                                        const sdp_session_t *sdp)
+{
+  TpsipMediaChannelPrivate *priv = TPSIP_MEDIA_CHANNEL_GET_PRIVATE (self);
+  const sdp_media_t *media;
+
+  for (media = sdp->sdp_media; media != NULL; media = media->m_next)
+    {
+      if (media->m_rejected || media->m_port == 0)
+        continue;
+
+      switch (media->m_type)
+        {
+        case sdp_media_audio:
+          priv->initial_audio = TRUE;
+          DEBUG("has initial audio");
+          break;
+        case sdp_media_video:
+          priv->initial_video = TRUE;
+          DEBUG("has initial video");
+          break;
+        default:
+          break;
+        }
+    }
+}
+
 static gboolean
 priv_nua_i_state_cb (TpsipMediaChannel *self,
                      const TpsipNuaEvent  *ev,
@@ -1176,13 +1352,30 @@ priv_nua_i_state_cb (TpsipMediaChannel *self,
 
   switch ((enum nua_callstate)ss_state)
     {
+    case nua_callstate_received:
+      if (tpsip_media_session_get_state (priv->session)
+          != TPSIP_MEDIA_SESSION_STATE_INVITE_RECEIVED)
+        break;
+      if (r_sdp != NULL)
+        priv_initial_media_properties_from_sdp (self, r_sdp);
+      g_signal_emit (self, signals[SIG_INCOMING_CALL], 0);
+      break;
     case nua_callstate_proceeding:
-      if (status == 180)
-        tpsip_media_channel_change_call_state (self, peer,
-                TP_CHANNEL_CALL_STATE_RINGING, 0);
-      else if (status == 182)
-        tpsip_media_channel_change_call_state (self, peer,
-                TP_CHANNEL_CALL_STATE_QUEUED, 0);
+      switch (status)
+        {
+          case 180:
+            tpsip_media_channel_change_call_state (self, peer,
+                    TP_CHANNEL_CALL_STATE_RINGING, 0);
+            break;
+          case 182:
+            tpsip_media_channel_change_call_state (self, peer,
+                    TP_CHANNEL_CALL_STATE_QUEUED, 0);
+            break;
+          case 183:
+            tpsip_media_channel_change_call_state (self, peer,
+                    TP_CHANNEL_CALL_STATE_IN_PROGRESS, 0);
+            break;
+        }
       break;
 
     case nua_callstate_completing:
@@ -1191,10 +1384,9 @@ priv_nua_i_state_cb (TpsipMediaChannel *self,
 
     case nua_callstate_ready:
 
-      /* Clear Ringing and Queued call states when the call is established */
+      /* Clear any pre-establishment call states */
       tpsip_media_channel_change_call_state (self, peer, 0,
-                TP_CHANNEL_CALL_STATE_RINGING |
-                TP_CHANNEL_CALL_STATE_QUEUED);
+          TPSIP_CHANNEL_CALL_STATE_PROCEEDING_MASK);
 
       if (status < 300)
         tpsip_media_session_accept (priv->session);
@@ -1203,7 +1395,7 @@ priv_nua_i_state_cb (TpsipMediaChannel *self,
       else
         {
           /* Was something wrong with our re-INVITE? We can't cope anyway. */
-          g_message ("can't handle non-fatal response %d %s", status, ev->text);
+          MESSAGE ("can't handle non-fatal response %d %s", status, ev->text);
           tpsip_media_session_terminate (priv->session);
         }
       break;
@@ -1561,7 +1753,7 @@ _tpsip_media_channel_add_member (GObject *iface,
       return TRUE;
     }
 
-  g_message ("unsupported member change requested for a media channel");
+  MESSAGE ("unsupported member change requested for a media channel");
 
   g_set_error (error, TP_ERRORS, TP_ERROR_NOT_AVAILABLE,
       "handle %u cannot be added in the current state", handle);
@@ -1710,7 +1902,14 @@ tpsip_media_channel_request_hold (TpSvcChannelInterfaceHold *iface,
 
   priv = TPSIP_MEDIA_CHANNEL_GET_PRIVATE (self);
 
-  if (priv->session != NULL)
+  if (priv->immutable_streams)
+    {
+      GError e = {TP_ERRORS, TP_ERROR_NOT_AVAILABLE,
+                  "Session modification disabled"};
+      dbus_g_method_return_error (context, &e);
+      return;
+    }
+  else if (priv->session != NULL)
     {
       tpsip_media_session_request_hold (priv->session, hold);
     }
@@ -1719,6 +1918,7 @@ tpsip_media_channel_request_hold (TpSvcChannelInterfaceHold *iface,
       GError e = {TP_ERRORS, TP_ERROR_NOT_AVAILABLE,
                   "The media session is not available"};
       dbus_g_method_return_error (context, &e);
+      return;
     }
 
   tp_svc_channel_interface_hold_return_from_request_hold (context);
