@@ -1,7 +1,7 @@
 /*
  * sip-media-session.c - Source for TpsipMediaSession
  * Copyright (C) 2005 Collabora Ltd.
- * Copyright (C) 2005-2008 Nokia Corporation
+ * Copyright (C) 2005-2010 Nokia Corporation
  *   @author Kai Vehmanen <first.surname@nokia.com>
  *   @author Mikhail Zabaluev <mikhail.zabaluev@nokia.com>
  *
@@ -77,8 +77,11 @@ enum
   PROP_PEER,
   PROP_HOLD_STATE,
   PROP_HOLD_STATE_REASON,
+  PROP_REMOTE_PTIME,
+  PROP_REMOTE_MAX_PTIME,
   PROP_RTCP_ENABLED,
   PROP_LOCAL_IP_ADDRESS,
+  PROP_STUN_SERVERS,
   LAST_PROPERTY
 };
 
@@ -124,6 +127,8 @@ struct _TpsipMediaSessionPrivate
   nua_handle_t *nua_op;                   /* see gobj. prop. 'nua-handle' */
   TpHandle peer;                          /* see gobj. prop. 'peer' */
   gchar *local_ip_address;                /* see gobj. prop. 'local-ip-address' */
+  gchar *remote_ptime;                    /* see gobj. prop. 'remote-ptime' */
+  gchar *remote_max_ptime;                /* see gobj. prop. 'remote-max-ptime' */
   gboolean rtcp_enabled;                  /* see gobj. prop. 'rtcp-enabled' */
   TpsipMediaSessionState state;           /* session state */
   TpLocalHoldState hold_state;         /* local hold state aggregated from stream directions */
@@ -160,9 +165,6 @@ tpsip_media_session_get_stream (TpsipMediaSession *self,
                               guint stream_id,
                               GError **error);
 
-static TpsipMediaStream* priv_create_media_stream (TpsipMediaSession *session,
-                                                 guint media_type,
-                                                 guint pending_send_flags);
 static void priv_request_response_step (TpsipMediaSession *session);
 static void priv_session_invite (TpsipMediaSession *session, gboolean reinvite);
 static void priv_local_media_changed (TpsipMediaSession *session);
@@ -229,12 +231,57 @@ static void tpsip_media_session_get_property (GObject    *object,
     case PROP_HOLD_STATE_REASON:
       g_value_set_uint (value, priv->hold_reason);
       break;
+    case PROP_REMOTE_PTIME:
+      g_value_set_string (value, priv->remote_ptime);
+      break;
+    case PROP_REMOTE_MAX_PTIME:
+      g_value_set_string (value, priv->remote_max_ptime);
+      break;
     case PROP_LOCAL_IP_ADDRESS:
       g_value_set_string (value, priv->local_ip_address);
       break;
     case PROP_RTCP_ENABLED:
       g_value_set_boolean (value, priv->rtcp_enabled);
       break;
+    case PROP_STUN_SERVERS:
+      {
+        /* TODO: should be able to get all entries from the DNS lookup(s).
+         * At the moment, rawudp ignores all servers except the first one. */
+        GPtrArray *servers;
+        gchar *stun_server = NULL;
+        guint stun_port = TPSIP_DEFAULT_STUN_PORT;
+
+        g_return_if_fail (priv->channel != NULL);
+
+        g_object_get (priv->channel,
+            "stun-server", &stun_server,
+            "stun-port", &stun_port,
+            NULL);
+
+        servers = g_ptr_array_new ();
+
+        if (stun_server != NULL)
+          {
+            GValue addr = { 0 };
+            const GType addr_type = TP_STRUCT_TYPE_SOCKET_ADDRESS_IP;
+
+            g_value_init (&addr, addr_type);
+            g_value_take_boxed (&addr,
+                dbus_g_type_specialized_construct (addr_type));
+
+            dbus_g_type_struct_set (&addr,
+                0, stun_server,
+                1, (guint16) stun_port,
+                G_MAXUINT);
+
+            g_ptr_array_add (servers, g_value_get_boxed (&addr));
+
+            g_free (stun_server);
+          }
+
+        g_value_take_boxed (value, servers);
+      }
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
       break;
@@ -337,6 +384,20 @@ tpsip_media_session_class_init (TpsipMediaSessionClass *klass)
       G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class, PROP_HOLD_STATE_REASON, param_spec);
 
+  param_spec = g_param_spec_string ("remote-ptime",
+      "a=ptime value of remote media session",
+      "Value of the a=ptime attribute of the remote media session, or NULL",
+      NULL,
+      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_REMOTE_PTIME, param_spec);
+
+  param_spec = g_param_spec_string ("remote-max-ptime",
+      "a=maxptime value of remote media session",
+      "Value of the a=maxptime attribute of the remote media session, or NULL",
+      NULL,
+      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_REMOTE_MAX_PTIME, param_spec);
+
   param_spec = g_param_spec_string ("local-ip-address", "Local IP address",
       "The local IP address preferred for media streams",
       NULL,
@@ -349,6 +410,12 @@ tpsip_media_session_class_init (TpsipMediaSessionClass *klass)
       G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class, PROP_RTCP_ENABLED, param_spec);
 
+  param_spec = g_param_spec_boxed ("stun-servers", "STUN servers",
+      "Array of IP address-port pairs for available STUN servers",
+      TP_ARRAY_TYPE_SOCKET_ADDRESS_IP_LIST,
+      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_STUN_SERVERS, param_spec);
+
   signals[SIG_STATE_CHANGED] =
     g_signal_new ("state-changed",
                   G_OBJECT_CLASS_TYPE (klass),
@@ -397,7 +464,7 @@ tpsip_media_session_finalize (GObject *object)
     TpsipMediaStream *stream = g_ptr_array_index (priv->streams, i);
     if (stream != NULL)
       {
-        g_warning ("stream %u (%p) left over, reaping", i, stream);
+        WARNING ("stream %u (%p) left over, reaping", i, stream);
         g_object_unref (stream);
       }
   }
@@ -410,6 +477,8 @@ tpsip_media_session_finalize (GObject *object)
   if (priv->backup_home != NULL)
     su_home_unref (priv->backup_home);
 
+  g_free (priv->remote_ptime);
+  g_free (priv->remote_max_ptime);
   g_free (priv->local_ip_address);
   g_free (priv->object_path);
 
@@ -548,24 +617,24 @@ priv_close_all_streams (TpsipMediaSession *session)
 }
 
 static void
-priv_apply_streams_pending_send (TpsipMediaSession *session,
-                                 guint pending_send_mask)
+priv_apply_streams_pending_direction (TpsipMediaSession *session,
+                                      guint pending_send_mask)
 {
   TpsipMediaSessionPrivate *priv = TPSIP_MEDIA_SESSION_GET_PRIVATE (session);
   TpsipMediaStream *stream;
   guint i;
 
   /* If there has been a local change pending a re-INVITE,
-   * leave pending remote send for the next transaction */
+   * suspend remote approval until the next transaction */
   if (priv->pending_offer)
     pending_send_mask &= ~(guint)TP_MEDIA_STREAM_PENDING_REMOTE_SEND;
 
-  /* Apply the local pending send flags where applicable */
+  /* Apply the pending direction changes */
   for (i = 0; i < priv->streams->len; i++)
     {
       stream = g_ptr_array_index(priv->streams, i);
       if (stream != NULL)
-        tpsip_media_stream_apply_pending_send (stream, pending_send_mask);
+        tpsip_media_stream_apply_pending_direction (stream, pending_send_mask);
     }
 }
 
@@ -600,8 +669,8 @@ tpsip_media_session_change_state (TpsipMediaSession *session,
       /* Apply any pending remote send after outgoing INVITEs.
        * We don't want automatic removal of pending local send after
        * responding to incoming re-INVITEs, however */
-      priv_apply_streams_pending_send (session,
-                                       TP_MEDIA_STREAM_PENDING_REMOTE_SEND);
+      priv_apply_streams_pending_direction (session,
+          TP_MEDIA_STREAM_PENDING_REMOTE_SEND);
       break;
     case TPSIP_MEDIA_SESSION_STATE_ENDED:
       priv_close_all_streams (session);
@@ -643,9 +712,8 @@ tpsip_media_session_debug (TpsipMediaSession *session,
 
   va_end (list);
 
-  tpsip_debug (DEBUG_FLAG, "SIP media session [%-17s]: %s",
-      session_states[priv->state],
-      buf);
+  DEBUG ("SIP media session [%-17s]: %s",
+      session_states[priv->state], buf);
 }
 #endif /* ENABLE_DEBUG */
 
@@ -836,9 +904,10 @@ gboolean tpsip_media_session_request_streams (TpsipMediaSession *session,
     guint media_type = g_array_index (media_types, guint, i);
     TpsipMediaStream *stream;
 
-    stream = priv_create_media_stream (session,
-                                       media_type,
-                                       TP_MEDIA_STREAM_PENDING_REMOTE_SEND);
+    stream = tpsip_media_session_add_stream (session,
+        media_type,
+        TP_MEDIA_STREAM_DIRECTION_BIDIRECTIONAL,
+        TRUE);
 
     if (stream == NULL)
       {
@@ -966,7 +1035,7 @@ priv_zap_event (TpsipMediaSession *self)
     {
       nua_event_data_t const *ev_data = nua_event_data (priv->saved_event);
       g_assert (ev_data != NULL);
-      g_warning ("zapping unhandled saved event '%s'", nua_event_name (ev_data->e_event));
+      WARNING ("zapping unhandled saved event '%s'", nua_event_name (ev_data->e_event));
       nua_destroy_event (priv->saved_event);
     }
 }
@@ -1022,9 +1091,9 @@ tpsip_media_session_accept (TpsipMediaSession *self)
   priv->accepted = TRUE;
 
   /* Apply the pending send flags */
-  priv_apply_streams_pending_send (self,
-                                   TP_MEDIA_STREAM_PENDING_LOCAL_SEND |
-                                   TP_MEDIA_STREAM_PENDING_REMOTE_SEND);
+  priv_apply_streams_pending_direction (self,
+      TP_MEDIA_STREAM_PENDING_LOCAL_SEND |
+      TP_MEDIA_STREAM_PENDING_REMOTE_SEND);
 
   /* Will change session state to active when streams are ready */
   priv_request_response_step (self);
@@ -1166,17 +1235,19 @@ priv_initiate_hold (TpsipMediaSession *self,
 
   if (hold)
     {
-      if (priv->hold_state == TP_LOCAL_HOLD_STATE_HELD)
+      if (priv->hold_state == TP_LOCAL_HOLD_STATE_HELD
+          || priv->hold_state == TP_LOCAL_HOLD_STATE_PENDING_HOLD)
         {
-          g_message ("redundant hold request");
+          MESSAGE ("redundant hold request");
           return;
         }
     }
   else
     {
-      if (priv->hold_state == TP_LOCAL_HOLD_STATE_UNHELD)
+      if (priv->hold_state == TP_LOCAL_HOLD_STATE_UNHELD
+          || priv->hold_state == TP_LOCAL_HOLD_STATE_PENDING_UNHOLD)
         {
-          g_message ("redundant unhold request");
+          MESSAGE ("redundant unhold request");
           return;
         }
     }
@@ -1185,18 +1256,9 @@ priv_initiate_hold (TpsipMediaSession *self,
   for (i = 0; i < priv->streams->len; i++)
     {
       stream = g_ptr_array_index(priv->streams, i);
-      if (stream != NULL)
-        {
-          gboolean stream_held = FALSE;
-          g_object_get (stream,
-                        "hold-state", &stream_held,
-                        NULL);
-          if ((!stream_held) != (!hold))
-            {
-              tp_svc_media_stream_handler_emit_set_stream_held (stream, hold);
-              stream_hold_requested = TRUE;
-            }
-        }
+      if (stream != NULL
+          && tpsip_media_stream_request_hold_state (stream, hold))
+        stream_hold_requested = TRUE;
     }
 
   if (stream_hold_requested)
@@ -1225,25 +1287,45 @@ priv_finalize_hold (TpsipMediaSession *self)
   guint hold_mask;
   guint unhold_mask;
   guint i;
+  gboolean held = FALSE;
 
   DEBUG("enter");
 
   switch (priv->hold_state)
     {
     case TP_LOCAL_HOLD_STATE_PENDING_HOLD:
+      held = TRUE;
+      break;
+    case TP_LOCAL_HOLD_STATE_PENDING_UNHOLD:
+      held = FALSE;
+      break;
+    default:
+      /* Streams changed state without request, signal this to the client.
+       * All streams should have the same hold state at this point,
+       * so just query one of them for the current hold state */
+      stream = NULL;
+      for (i = 0; i < priv->streams->len; i++)
+        {
+          stream = g_ptr_array_index(priv->streams, i);
+          if (stream != NULL)
+            break;
+        }
+      g_return_if_fail (stream != NULL);
+
+      g_object_get (stream, "hold-state", &held, NULL);
+    }
+
+  if (held)
+    {
       final_hold_state = TP_LOCAL_HOLD_STATE_HELD;
       hold_mask = TP_MEDIA_STREAM_DIRECTION_SEND;
       unhold_mask = 0;
-      break;
-    case TP_LOCAL_HOLD_STATE_PENDING_UNHOLD:
+    }
+  else
+    {
       final_hold_state = TP_LOCAL_HOLD_STATE_UNHELD;
       hold_mask = TP_MEDIA_STREAM_DIRECTION_BIDIRECTIONAL;
       unhold_mask = TP_MEDIA_STREAM_DIRECTION_RECEIVE;
-      break;
-    default:
-      /* This internal function must not be called in final hold states */
-      g_assert_not_reached ();
-      return;
     }
 
   priv->hold_state = final_hold_state;
@@ -1397,6 +1479,18 @@ priv_local_media_changed (TpsipMediaSession *session)
       priv->pending_offer = TRUE;
       break;
     case TPSIP_MEDIA_SESSION_STATE_ACTIVE:
+      /* Check if we are allowed to send re-INVITES */
+      {
+        gboolean immutable_streams = FALSE;
+        g_object_get (priv->channel,
+            "immutable-streams", &immutable_streams,
+            NULL);
+        if (immutable_streams) {
+          g_message ("sending of a local media update disabled by parameter 'immutable-streams'");
+          break;
+        }
+      }
+      /* Fall through to the next case */
     case TPSIP_MEDIA_SESSION_STATE_REINVITE_PENDING:
       if (priv->local_non_ready == 0)
         priv_session_invite (session, TRUE);
@@ -1450,22 +1544,41 @@ priv_update_remote_hold (TpsipMediaSession *session)
                                            TP_CHANNEL_CALL_STATE_HELD);
 }
 
+gchar *
+tpsip_sdp_get_string_attribute (const sdp_attribute_t *attrs, const char *name)
+{
+  sdp_attribute_t *attr;
+
+  attr = sdp_attribute_find (attrs, name);
+  if (attr == NULL)
+    return NULL;
+
+  return g_strdup (attr->a_value);
+}
+
 static gboolean
 priv_update_remote_media (TpsipMediaSession *session, gboolean authoritative)
 {
   TpsipMediaSessionPrivate *priv = TPSIP_MEDIA_SESSION_GET_PRIVATE (session);
+  const sdp_session_t *sdp = priv->remote_sdp;
   const sdp_media_t *media;
   gboolean has_supported_media = FALSE;
   guint direction_up_mask;
   guint pending_send_mask;
   guint i;
 
-  g_return_val_if_fail (priv->remote_sdp != NULL, FALSE);
+  g_return_val_if_fail (sdp != NULL, FALSE);
+
+  /* Update the session-wide parameters
+   * before updating streams' media */
+
+  priv->remote_ptime     = tpsip_sdp_get_string_attribute (
+      sdp->sdp_attributes, "ptime");
+  priv->remote_max_ptime = tpsip_sdp_get_string_attribute (
+      sdp->sdp_attributes, "maxptime");
 
-  /* Update the session-wide RTCP enable flag
-   * before updating stream media */
   priv->rtcp_enabled = !tpsip_sdp_rtcp_bandwidth_throttled (
-                                priv->remote_sdp->sdp_bandwidths);
+                                sdp->sdp_bandwidths);
 
   /*
    * Do not allow:
@@ -1488,7 +1601,7 @@ priv_update_remote_media (TpsipMediaSession *session, gboolean authoritative)
   if (priv->pending_offer)
     pending_send_mask |= TP_MEDIA_STREAM_PENDING_REMOTE_SEND;
 
-  media = priv->remote_sdp->sdp_media;
+  media = sdp->sdp_media;
 
   /* note: for each session, we maintain an ordered list of
    *       streams (SDP m-lines) which are matched 1:1 to
@@ -1502,10 +1615,11 @@ priv_update_remote_media (TpsipMediaSession *session, gboolean authoritative)
       media_type = tpsip_tp_media_type (media->m_type);
 
       if (i >= priv->streams->len)
-        stream = priv_create_media_stream (
+        stream = tpsip_media_session_add_stream (
                         session,
                         media_type,
-                        TP_MEDIA_STREAM_PENDING_LOCAL_SEND);
+                        tpsip_media_stream_direction_from_remote_media (media),
+                        FALSE);
       else
         stream = g_ptr_array_index(priv->streams, i);
 
@@ -1522,7 +1636,7 @@ priv_update_remote_media (TpsipMediaSession *session, gboolean authoritative)
       else if (tpsip_media_stream_get_media_type (stream) != media_type)
         {
           /* XXX: close this stream and create a new one in its place? */
-          g_warning ("The peer has changed the media type, don't know what to do");
+          WARNING ("The peer has changed the media type, don't know what to do");
         }
       else if (tpsip_media_stream_set_remote_media (stream,
                                                     media,
@@ -1559,7 +1673,7 @@ priv_update_remote_media (TpsipMediaSession *session, gboolean authoritative)
           stream = g_ptr_array_index(priv->streams, i);
           if (stream != NULL)
             {
-              g_message ("closing a mismatched stream %u", i);
+              MESSAGE ("closing a mismatched stream %u", i);
               tpsip_media_stream_close (stream);
             }
         }
@@ -1692,7 +1806,7 @@ priv_session_invite (TpsipMediaSession *session, gboolean reinvite)
                         : TPSIP_MEDIA_SESSION_STATE_INVITE_SENT);
     }
   else
-    g_warning ("cannot send a valid SDP offer, are there no streams?");
+    WARNING ("cannot send a valid SDP offer, are there no streams?");
 
   g_string_free (user_sdp, TRUE);
 }
@@ -1733,7 +1847,7 @@ priv_session_respond (TpsipMediaSession *session)
     }
   else
     {
-      g_warning ("cannot respond with a valid SDP answer, were all streams closed?");
+      WARNING ("cannot respond with a valid SDP answer, were all streams closed?");
 
       priv_session_rollback (session);
     }
@@ -1937,7 +2051,7 @@ priv_stream_hold_state_cb (TpsipMediaStream *stream,
       hold = FALSE;
       break;
     default:
-      g_message ("unexpected hold state change from a stream");
+      MESSAGE ("unexpected hold state change from a stream");
 
       /* Try to follow the changes and report the resulting hold state */
       g_object_get (stream, "hold-state", &hold, NULL);
@@ -1972,37 +2086,35 @@ priv_stream_unhold_failure_cb (TpsipMediaStream *stream,
                       TP_LOCAL_HOLD_STATE_REASON_RESOURCE_NOT_AVAILABLE);
 }
 
-static TpsipMediaStream*
-priv_create_media_stream (TpsipMediaSession *self,
-                          guint media_type,
-                          guint pending_send_flags)
+TpsipMediaStream*
+tpsip_media_session_add_stream (TpsipMediaSession *self,
+                                guint media_type,
+                                TpMediaStreamDirection direction,
+                                gboolean created_locally)
 {
   TpsipMediaSessionPrivate *priv = TPSIP_MEDIA_SESSION_GET_PRIVATE (self);
-  gchar *object_path;
   TpsipMediaStream *stream = NULL;
-  guint stream_id;
-  TpMediaStreamDirection direction;
 
   DEBUG ("enter");
 
   if (tpsip_media_session_supports_media_type (media_type)) {
+    guint stream_id;
+    gchar *object_path;
+    guint pending_send_flags;
 
     stream_id = priv->streams->len;
     object_path = g_strdup_printf ("%s/MediaStream%u",
                                    priv->object_path,
                                    stream_id);
+    pending_send_flags = created_locally
+        ? TP_MEDIA_STREAM_PENDING_REMOTE_SEND
+        : TP_MEDIA_STREAM_PENDING_LOCAL_SEND;
+
+    if (!created_locally)
+      direction &= ~TP_MEDIA_STREAM_DIRECTION_SEND;
+
     if (tpsip_media_session_is_local_hold_ongoing (self))
-      {
-        direction = (pending_send_flags == 0)
-                ? TP_MEDIA_STREAM_DIRECTION_SEND
-                : TP_MEDIA_STREAM_DIRECTION_NONE;
-      }
-    else
-      {
-        direction = (pending_send_flags == 0)
-                ? TP_MEDIA_STREAM_DIRECTION_BIDIRECTIONAL
-                : TP_MEDIA_STREAM_DIRECTION_RECEIVE;
-      }
+      direction &= ~TP_MEDIA_STREAM_DIRECTION_RECEIVE;
 
     stream = g_object_new (TPSIP_TYPE_MEDIA_STREAM,
 			   "media-session", self,
@@ -2011,6 +2123,7 @@ priv_create_media_stream (TpsipMediaSession *self,
 			   "id", stream_id,
                            "direction", direction,
                            "pending-send-flags", pending_send_flags,
+                           "created-locally", created_locally,
 			   NULL);
 
     g_free (object_path);
@@ -2050,11 +2163,15 @@ priv_create_media_stream (TpsipMediaSession *self,
                                                           stream_id,
                                                           priv->peer,
                                                           media_type);
-    tp_svc_channel_type_streamed_media_emit_stream_direction_changed (
-        priv->channel,
-        stream_id,
-        direction,
-        pending_send_flags);
+    if (direction != TP_MEDIA_STREAM_DIRECTION_RECEIVE
+        || pending_send_flags != TP_MEDIA_STREAM_PENDING_LOCAL_SEND)
+      {
+        tp_svc_channel_type_streamed_media_emit_stream_direction_changed (
+            priv->channel,
+            stream_id,
+            direction,
+            pending_send_flags);
+      }
   }
 
   /* note: we add an entry even for unsupported media types */
